from datetime import date
import json

from django.apps import apps
from django.core.mail import EmailMessage
from django.urls import reverse
from django.utils.translation import ugettext as _
from django.shortcuts import redirect, render
from django_tables2 import RequestConfig

from .forms import EventRegisterForm
from .models import TeckidsProject
from .tables import EventsTable
from .util import form_to_text_table

TeckidsPerson = apps.get_model('ticdesk_account', 'TeckidsPerson')

def upcoming_events(request):
    context = {}

    # Get all upcoming events
    now = date.today().strftime('%Y-%m-%d')
    events = TeckidsProject.objects.filter(date_event__gte=now)

    # Build table
    events_table = EventsTable(events)
    RequestConfig(request).configure(events_table)
    context['events_table'] = events_table

    return render(request, 'ticdesk_events/events.html', context)

def register_event(request, cn):
    context = {}
    
    # Only authenticated users an register for events
    if not request.user.is_authenticated():
        return redirect(reverse('ticdesk_account:login') + '?next=%s' % request.path)

    # Get current person and event
    current_person = TeckidsPerson.objects.get(uid=request.user.username)
    event = TeckidsProject.objects.get(cn=cn)
    context['event'] = event

    # Prepare the form
    initial = {
        'cn': current_person.cn,
        'date_of_birth': current_person.date_of_birth,
        'mail': current_person.mail,
        'street': current_person.home_postal_street,
        'postal_code': current_person.home_postal_code,
        'place': current_person.home_postal_place,
        'mobile': current_person.mobile,
        'school': current_person.o,
        'school_place': current_person.l,
        'school_class': current_person.ou,
        'channel': 'none'
    }
    register_form = EventRegisterForm(initial=initial)

    # Produce error if registration is not possible
    if not event.can_register:
        context['error'] = _('Die Anmeldung ist nicht mehr möglich!')
        return render(request, 'ticdesk_events/register_event.html', context)

    # Check whether person is already a member of the event
    if current_person.dn in event.members:
        context['error'] = _('Du bist bereits angemeldet.')
        return render(request, 'ticdesk_events/register_event.html', context)

    if request.method == 'POST':
        register_form = EventRegisterForm(request.POST, initial=initial)
        if register_form.is_valid():
            # Add the current person to the event
            event.add_member(current_person)
            event.save()
            
            # Update person information if changed
            if ('school' in register_form.changed_data or
                'school_class' in register_form.changed_data or
                'school_place' in register_form.changed_data or
                'mobile' in register_form.changed_data
            ):
                current_person.o = register_form.cleaned_data['school']
                current_person.ou = register_form.cleaned_data['school_class']
                current_person.l = register_form.cleaned_data['school_place']
                current_person.mobile = register_form.cleaned_data['mobile']
                current_person.save()

        # Produce e-mail to registration queue
        message = EmailMessage()
        message.reply_to = (current_person.mail, register_form.cleaned_data['guardian_mail'])
        message.to = ['anmeldung@teckids.org',]
        message.subject = _('Anmeldung: %s von %s') % (event.display_name, current_person.cn)
        message.extra_headers = {
            'X-OTRS-DynamicField-TeckidsEvent': event.display_name,
            'X-OTRS-CustomerUser': current_person.uid
        }
        message.body = ''
        if ('street' in register_form.changed_data or
            'postal_code' in register_form.changed_data or
            'place' in register_form.changed_data
        ):
            message.body += '***** ' + _('Geänderte Anschrift') + ' *****\n\n'
        message.body += form_to_text_table(register_form, 78)

        # Attach raw form data as attachment
        message.attach('register_form.json', json.dumps(register_form.cleaned_data, indent=4, default=str), 'application/json')

        # Send message
        message.send()

        # Set success
        context['success'] = True

    context['register_form'] = register_form

    return render(request, 'ticdesk_events/register_event.html', context)
